@@ -30,6 +30,7 @@
 #include <unistd.h>
 
 #include <algorithm>
+#include <charconv>
 #include <numeric>
 #include <regex>
 #include <lomiri/util/GlibMemory.h>
@@ -939,9 +940,49 @@ SystemD::UnitInfo SystemD::parseUnit(const std::string& unit) const
     return {match[2].str(), match[1].str(), match[3].str()};
 }
 
+//  https://github.com/coreos/go-systemd/blob/v22.5.0/unit/escape.go#L36
+//
+//	Replace the following characters with `\x%x`: Leading `.`,
+//	 `-`, `\`, and anything not in this set: `:-_.\[0-9a-zA-Z]`
+//	Replace '/' with '-'.
+//
+// (Note that we're dealing with component, so the "leading `.`" part
+// doesn't apply. Also we're allowing additionally '-' because we're not going
+// to unescape the name back.)
+static std::string escapeUnitComponent(const std::string & unescaped)
+{
+    const std::string allowed = ":-_.abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
+
+    std::string escaped;
+    // Escaped string can only be longer, so reserve at least unescaped length.
+    escaped.reserve(unescaped.length());
+
+    for (size_t i = 0; i < unescaped.length(); i++) {
+        char c = unescaped[i];
+
+        if (c == '/') {
+            escaped += '-';
+        } else if (allowed.find(c) == std::string::npos) {
+            // https://stackoverflow.com/a/72132640
+            char buf[3] = { 0 };
+            std::to_chars(std::begin(buf), std::end(buf), (unsigned int)c, 16);
+
+            escaped += "\\x";
+            escaped += buf;
+        } else {
+            escaped += c;
+        }
+    }
+
+    return escaped;
+}
+
 std::string SystemD::unitName(const SystemD::UnitInfo& info) const
 {
-    return std::string{"lomiri-app-launch--"} + info.job + "--" + info.appid + "--" + info.inst + ".service";
+    return std::string{"lomiri-app-launch--"} + info.job
+                + "--" + escapeUnitComponent(info.appid)
+                + "--" + info.inst
+                + ".service";
 }
 
 std::string SystemD::unitPath(const SystemD::UnitInfo& info)
